package org.openxma.dsl.generator.component;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.mwe.core.WorkflowContext;
import org.eclipse.emf.mwe.core.issues.Issues;
import org.eclipse.emf.mwe.core.monitor.ProgressMonitor;
import org.eclipse.emf.mwe.utils.Writer;

import at.spardat.xma.guidesign.XMAComponent;

/**
 * Writer for a XMAComponent.
 * 
 * It is a wrapper around XmiWriter and has the purpose to determine the name
 * and location of the xma-file using information of the XMAComponent.
 * 
 */
public class XMAWriter extends Writer {
	
	public final static String XMA_MODEL_REL_PATH_KEY = "XMAModelRelativePath";	
	
	private String platformUri;
	private String projectName;
	private String sourceFolder;
	private String editorFileFolder; // May be used to determine the sourceFolder if not specified explicit		
	private String modelFile;
	/**
	 * Sets the name of the model file.
	 * 
	 * @param modelFile
	 *            name of file
	 */
	public void setModelFile(final String modelFile) {
		this.modelFile = modelFile;
	}

	/**
	 * Returns the name of the model file.
	 * 
	 * @return name of model file
	 */
	public String getModelFile() {
		return modelFile;
	}

	/**
	 * Default constructor 
	 */
	public XMAWriter() {
		super();
	}	

	/**
	 * Determine the source folder name.
	 * 
	 * @param packagePath
	 * @return
	 */
	protected String determineSourceFolder(String packagePath) {		
		if (sourceFolder != null) return sourceFolder;
		if (editorFileFolder != null) {			
			int firstPackageCharacterIndex=-1;
			if (editorFileFolder.endsWith(packagePath)) {
				firstPackageCharacterIndex = editorFileFolder.indexOf(packagePath)-1;							
			} else {
				firstPackageCharacterIndex = editorFileFolder.indexOf("/");
			}			
			if (firstPackageCharacterIndex>0) {
				return editorFileFolder.substring(0,firstPackageCharacterIndex);
			}
		}		
		// Default value 
		return "src";
	}
	
	/**
	 * Write the XMAComponent which is stored in the modelslot.
	 * 
	 * Therefore the modelslot must not be empty and the content must be a XMAComponent.
	 * In case the XMAComponent exists already it is be cloned before it is written. 
	 */
	public void invokeInternal(final WorkflowContext ctx, final ProgressMonitor monitor, final Issues issues) {		
		final Object slotContent = ctx.get(getModelSlot());
		if (slotContent == null) {
			issues.addError(this, "slot '" + getModelSlot() + "' is empty.");
			return;
		}		
		if (!(slotContent instanceof XMAComponent)) {
			issues.addError(this, "slot '" + getModelSlot() + "' contains something else then a XMAComponent",
					slotContent, null, null);
			System.out.println("XMAComponent not found in input slot");				
			return;
		}			
		XMAComponent comp = (XMAComponent)slotContent;		
		String relativeModelPath = getRelativeModelPathForXMAComponent(comp);
		URI fileURI = URI.createFileURI(platformUri+"/"+projectName+"/"+relativeModelPath);
		setUri(fileURI.toString());
		// In the case the resource for XMAComponent exists already copy it to be consistent in case of a error
		if (comp.eResource() != null) {			
			setCloneSlotContents(true);
		}
		checkConfiguration(issues);			
		super.invokeInternal(ctx, monitor, issues);
		ctx.set(XMA_MODEL_REL_PATH_KEY, relativeModelPath);							
	}
	
	protected String getRelativeModelPathForXMAComponent(XMAComponent comp) {
		String packagePath = comp.getNamPackageAsPath();
		String componentName = comp.getNamClass();		
		String sourceFolder = determineSourceFolder(packagePath);
		if (sourceFolder.endsWith("/")==false) {
			sourceFolder+="/";
		}			
		return sourceFolder+packagePath+"/"+componentName+".xma";				
	}

	
	public void setEditorFileFolder(String sourceFileFolder) {
		this.editorFileFolder = sourceFileFolder;
	}

	public String getEditorFileFolder() {
		return editorFileFolder;
	}
	
	public String getSourceFolder() {
		return sourceFolder;		
	}

	public void setSourceFolder(String sourceFolder) {
		this.sourceFolder = sourceFolder;
	}
	
	public String getPlatformUri() {
		return platformUri;
	}

	public void setPlatformUri(String platformUri) {
		this.platformUri = platformUri;
	}

	public String getProjectName() {
		return projectName;
	}

	public void setProjectName(String projectName) {
		this.projectName = projectName;
	}	
}
