package org.openxma.dsl.generator.helper;

import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.openxma.dsl.common.extensions.EObjectPropertiesAdapter;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.dom.model.DataView;

public class DataViewExtension {

    /**
     * sets a property on the <code>ResourceSet</code> indicating that a DtoAssemblerFactory has been generated for the
     * namespace of the model
     * 
     * @param model
     *            the model (dml file) whose namespace a DtoAssemblerFactory will be generated for
     * @return the name of the model namespace
     */
    public static String addGeneratedNameSpaceName(Model model) {
        ResourceSet resourceSet = model.eResource().getResourceSet();
        EObjectPropertiesAdapter.setProperty(resourceSet, model.getName(), model.getName());
        return model.getName();
    }

    /**
     * @param model
     *            the model (dml file) whose namespace is asked whether already a DtoAssemblerFactory has been generated
     *            for
     * @return <code>Boolean.TRUE</code> if a DtoAssemblerFactory has been generated already for the namespace of the
     *         model, otherwise <code>Boolean.FALSE</code>
     */
    public static Boolean isNameSpaceGenerated(Model model) {
        ResourceSet resourceSet = model.eResource().getResourceSet();
        return EObjectPropertiesAdapter.hasProperty(resourceSet, model.getName());
    }

    public static Set<DataView> getNameSpaceDataViewsFromModel(Model model) {
        if (null == model)
            return getSortByDataViewNameAscSetInstance();

        return extractNameSpaceDataViews(model.eResource().getResourceSet().getResources().iterator(), model.getName());
    }

    private static Set<DataView> getSortByDataViewNameAscSetInstance() {
        return new TreeSet<DataView>(new Comparator<DataView>() {
            public int compare(DataView dv1, DataView dv2) {
                return dv1.getName().compareTo(dv2.getName());
            }
        });
    }

    private static Set<DataView> extractNameSpaceDataViews(Iterator<Resource> iter, String nameSpace) {
        Set<DataView> result = getSortByDataViewNameAscSetInstance();

        while (iter.hasNext()) {
            Model nextModel = (Model) iter.next().getEObject("/");

            if (nextModel.getName().equals(nameSpace))
                extractNameSpaceDataViews(nextModel, result);
        }

        return result;
    }

    private static Set<DataView> extractNameSpaceDataViews(Model nameSpace, Set<DataView> result) {
        for (TreeIterator<EObject> eAllContents = nameSpace.eAllContents(); eAllContents.hasNext();) {
            EObject next = eAllContents.next();
            if (next instanceof DataView)
                extractNameSpaceDataViews((DataView) next, result, nameSpace.getName());
        }

        return result;
    }

    private static void extractNameSpaceDataViews(DataView dv, Set<DataView> nameSpaceDataViews, String nameSpaceName) {
        if (null != dv.getSuperType())
            extractNameSpaceDataViews(dv.getSuperType(), nameSpaceDataViews, nameSpaceName);

        // although we derive DataViews from the selected namespace, parent datasources could
        // come from other namespaces too: exclude them with the following check agains the name of the namespace
        if (((Model) dv.eContainer()).getName().equals(nameSpaceName))
            nameSpaceDataViews.add(dv);
    }

}
